# Hand-maintained public namespace. Do not add a roxygen2-generated header:
# DESCRIPTION disables the namespace roclet, and export changes must be made
# here deliberately with matching api-surface coverage.
useDynLib(eigencore, .registration = TRUE)
importFrom(methods, is)
importFrom(stats, residuals)
importFrom(utils, modifyList)
importFrom(utils, tail)
export(adjoint)
export(alpha_beta)
export(as_operator)
export(auto)
export(backward_error)
export(both_ends)
export(certificate)
export(center)
export(check_adjoint)
export(compose)
export(crossprod_operator)
export(diagnostics)
export(eigen_problem)
export(eig_full)
export(eig_partial)
export(eigs)
export(eigs_sym)
export(euclidean)
export(general)
export(generalized_schur)
export(generalized_svd)
export(golub_kahan)
export(hermitian)
export(lanczos)
export(largest)
export(largest_imaginary)
export(largest_magnitude)
export(largest_real)
export(left_vectors)
export(linear_operator)
export(lobpcg)
export(nearest)
export(plan_solver)
export(randomized)
export(right_vectors)
export(scale_cols)
export(scale_rows)
export(shift_invert)
export(shifted_cholesky_preconditioner)
export(shifted_diagonal_preconditioner)
export(shifted_tridiagonal_preconditioner)
export(smallest)
export(smallest_imaginary)
export(smallest_magnitude)
export(smallest_real)
export(symmetric_operator)
export(svd_partial)
export(svd_problem)
export(svds)
export(values)
export(vectors)
S3method(adjoint, eigencore_operator)
S3method(as_operator, default)
S3method(as_operator, eigencore_operator)
S3method(as_operator, matrix)
S3method(plan_solver, eigencore_eigen_problem)
S3method(plan_solver, eigencore_svd_problem)
S3method(print, eigencore_certificate)
S3method(print, eigencore_benchmark)
S3method(print, eigencore_eigen_result)
S3method(print, eigencore_gsvd_result)
S3method(print, eigencore_operator)
S3method(print, eigencore_plan)
S3method(print, eigencore_svd_result)
S3method(print, eigencore_validation)
S3method(residuals, eigencore_certificate)
S3method(residuals, eigencore_eigen_result)
S3method(residuals, eigencore_svd_result)
S3method(solve, eigencore_eigen_problem)
S3method(solve, eigencore_svd_problem)
